!
! Module to solve 1D Helmholtz equation using tridiagonal solver
!
! R. Orvedahl 9-17-2014

module helmholtz

   use variables
   use data_types
   use linear_algebra

   implicit none

   contains

   !----------------------------------------------------------------------
   ! solve 1D Constant Coefficient Helmholtz Equation:
   !      (a - b d2/dx2) phi = rhs
   !----------------------------------------------------------------------
   subroutine helmholtz_cc(n, dx, a, b, rhs, bc, bdry, phi)

      integer, intent(in) :: n, bdry
      real(kind=dpt), intent(in) :: a, b, rhs(0:n+1), dx, bc(2)
      real(kind=dpt), intent(inout) :: phi(0:n+1)

      real(kind=dpt) :: acoeff(n), bcoeff(n), ccoeff(n), rhstmp(0:n+1)
      real(kind=dpt) :: phi_l, phi_u

      ! unpack BCs
      phi_l = bc(1)
      phi_u = bc(2)

      ! system is in form:
      !      a(i)*phi(i-1) + b(i)*phi(i) + c(i)*phi(i+1) = dx**2*rhs(i) 
      ! for Helmholtz:
      !      acoeff(i) = -b
      !      bcoeff(i) = 2*b + dx**2*a
      !      ccoeff(i) = -b
      !      rhs(:) = dx**2*rhs(:) <-- must modify the rhs from discretization
      !
      ! modification for BC:
      !
      !    bdry = 0 (Dirichlet on both boundaries)
      !       rhs(lo) = rhs(lo) - two*acoeff(lo)*phi(lbdry)
      !       rhs(hi) = rhs(hi) - two*ccoeff(hi)*phi(ubdry)
      !       bcoeff(lo) = bcoeff(lo) - acoeff(lo)
      !       bcoeff(hi) = bcoeff(hi) - ccoeff(hi)
      !    bdry = 1 (Neumann on both boundaries)
      !       rhs(lo) = rhs(lo) - two*acoeff(lo)*phi(lbdry)
      !       rhs(hi) = rhs(hi) - two*ccoeff(hi)*phi(ubdry)
      !       bcoeff(lo) = bcoeff(lo) - acoeff(lo)
      !       bcoeff(hi) = bcoeff(hi) - ccoeff(hi)
      !    bdry = 2 (Dirichlet lower, Neumann Upper)
      !       rhs(lo) = rhs(lo) - two*acoeff(lo)*phi(lbdry)
      !       rhs(hi) = rhs(hi) - two*ccoeff(hi)*phi(ubdry)
      !       bcoeff(lo) = bcoeff(lo) - acoeff(lo)
      !       bcoeff(hi) = bcoeff(hi) - ccoeff(hi)
      !    bdry = 3 (Neumann lower, Dirichlet Upper)
      !       rhs(lo) = rhs(lo) - two*acoeff(lo)*phi(lbdry)
      !       rhs(hi) = rhs(hi) - two*ccoeff(hi)*phi(ubdry)
      !       bcoeff(lo) = bcoeff(lo) - acoeff(lo)
      !       bcoeff(hi) = bcoeff(hi) - ccoeff(hi)

      acoeff(:) = -b
      bcoeff(:) = two*b + a*dx**2
      ccoeff(:) = -b
      rhstmp = dx**2*rhs

      if (bdry == 0) then ! Dirichlet on lower and upper
          bcoeff(1) = bcoeff(1) - acoeff(1)
          rhstmp(1) = rhstmp(1) - two*acoeff(1)*phi_l
          bcoeff(n) = bcoeff(n) - ccoeff(n)
          rhstmp(n) = rhstmp(n) - two*ccoeff(n)*phi_u

      elseif (bdry == 1) then ! Neumann on lower and upper
          bcoeff(1) = bcoeff(1) + acoeff(1)
          rhstmp(1) = rhstmp(1) + acoeff(1)*dx*phi_l
          bcoeff(n) = bcoeff(n) + ccoeff(n)
          rhstmp(n) = rhstmp(n) - ccoeff(n)*dx*phi_u

      elseif (bdry == 2) then ! Dirichlet on lower, Neumann on upper
          bcoeff(1) = bcoeff(1) - acoeff(1)
          rhstmp(1) = rhstmp(1) - two*acoeff(1)*phi_l
          bcoeff(n) = bcoeff(n) + ccoeff(n)
          rhstmp(n) = rhstmp(n) - ccoeff(n)*dx*phi_u

      elseif (bdry == 3) then ! Neumann on lower, Dirichlet on upper
          bcoeff(1) = bcoeff(1) + acoeff(1)
          rhstmp(1) = rhstmp(1) + acoeff(1)*dx*phi_l
          bcoeff(n) = bcoeff(n) - ccoeff(n)
          rhstmp(n) = rhstmp(n) - two*ccoeff(n)*phi_u

      else
          write(*,*)
          write(*,*) "Error: unsupported boundary conditions, bdry: ",bdry
          write(*,*)
          stop
      endif

      ! solve system
      call tridiagonal(n, acoeff, bcoeff, ccoeff, rhstmp(1:n), phi(1:n))

      return

   end subroutine helmholtz_cc

end module helmholtz

