!
! program to solve constant coefficient helmholtz equation
!
! the computational grid is assumed to be cell-centered 
! finite difference or equivalently finite volume
!
!      a*phi - b*phi'' = rhs
!      along with 2 boundary conditions on phi
!
! R. Orvedahl (9-17-2014)

program test_helmholtz

   use data_types
   use variables
   use helmholtz

   implicit none

   integer, parameter :: nx = 128

   real(kind=dpt), allocatable, dimension(:) :: phi, rhs, phiexact, x

   real(kind=dpt) :: dx, a, b, l2, bc(2)
   integer :: i, bdry

   dx = one / real(nx,kind=dpt)

   ! 0 and nx+1 indicies are ghost cells to enforce the boundary conditions
   allocate(x(0:nx+1), phi(0:nx+1), rhs(0:nx+1), phiexact(0:nx+1))

   ! cell-centered finite difference grid (or equivalently finite volume)
   do i=0,nx+1
      x(i) = zero + dx*(real(i,kind=dpt) - half)
   enddo

   ! test helmholtz code with:
   !       a*phi - b*phi'' = f
   !       f = a*x*sin(pi*x) - a*x**2*cos(pi*x) -
   !           b*(2*pi+pi**2*x**2-2)*cos(pi*x) - b*(4*pi-pi**2)*x*sin(pi*x)
   !       phi(0) = 0
   !       phi(1) = 1
   !       solution:
   !         phi = x*sin(pi*x) - x**2*cos(pi*x)
   a = 2.0
   b = 5.0
   phiexact = x*sin(pi*x) - x**2*cos(pi*x)
   rhs = a*x*sin(pi*x) - a*x**2*cos(pi*x) - &
              b*(2*pi+pi**2*x**2-2)*cos(pi*x) - b*(4*pi-pi**2)*x*sin(pi*x)

   bdry = 0      ! defines the type of boundary condition
   bc(1) = zero  ! store the boundary values
   bc(2) = one

   ! solve for phi
   call helmholtz_cc(nx, dx, a, b, rhs, bc, bdry, phi)

   ! fill Dirichlet boundary conditions on both boundaries
   phi(0) = two*zero - phi(1)
   phi(nx+1) = two*one - phi(nx)

   print *,''
   print *, 'Test Helmholtz'

   ! calculate the L2 norm: sqrt( dx * sum((x(i) - xtrue(i))**2) )
   l2 = zero
   do i=1,nx
      l2  = l2 + (phi(i) - phiexact(i))**2
   enddo
   ! the dx is to make it grid space invariant
   l2  = sqrt(l2*dx)

   print *,''
   print *,'   L2 norm:',l2
   print *,''
   print *,'   nx:',nx
   print *,'   dx:',dx
   print *,''

   deallocate(phi, rhs, phiexact, x)

end program test_helmholtz

